/**
 * Auto save functionality
 * @param  {object} editor   The editor reference
 * @param  {string} pageID   The content object id
 * @param  {string} ddMenuID The menu ID
 * @param  {object} options  The override options such as: storeMax, timeout
 * @return {object}          Returns the auto save closure
 */
import { decompressFromUTF16, compressToUTF16 } from "lz-string";
import { DateTime } from "luxon";

const autoSavePrefix = "autosave_";

window.resetAutoSave = function(){
	if ( !Modernizr.localstorage ) return;
	let saveStoreKey 	= autoSavePrefix + window.location;
	let saved = localStorage.getItem( saveStoreKey );
	if ( saved ){
		JSON.parse( saved )
			.forEach( ( entry, index ) => {
				localStorage.removeItem( entry );
			} );
		localStorage.setItem( saveStoreKey, "[]" );
	}

};

window.autoSave = function( editor, pageID, ddMenuID, options ){
	// Verify local storage, else disable feature
	if ( !Modernizr.localstorage ){
		$( "#" + ddMenuID ).find( ".autoSaveBtn" )
			.html( "Auto Save Unavailable" )
			.addClass( "disabled" );
		return false;
	}

	// Setup defaults and global options
	let defaults 		= { storeMax: 10, timeout: 4000 };
	let opts 			= $.extend( {}, defaults, options || {} );
	let editorID 		= editor.attr( "id" );
	// Retrieve the actual editor driver implementation using ContentBox JS Interface Method
	let saveStoreKey 	= autoSavePrefix + window.location;
	let timer 			= 0, savingActive = false;

	// Setup SavesStore
	if ( !localStorage.getItem( saveStoreKey ) ){
		localStorage.setItem( saveStoreKey, "[]" );
	}
	let saveStore = JSON.parse( localStorage.getItem( saveStoreKey ) );

	/**
	 * Remove old saves
	 * @param  {Function} callback The callback function
	 */
	let removeOldSaves = function( callback ){
		let overMax = saveStore.length - opts.storeMax;
		for ( let i = 0; i < overMax; i++ ){
		  localStorage.removeItem( saveStore[ i ] );
		  saveStore.splice( i, 1 );
		}
		if ( callback ){
			callback();
		}
	};

	/**
	 * Add to saved storage
	 * @param {string} saveKey The save storage key
	 */
	let addToStore = function( saveKey ){
		saveStore.push( saveKey );
		if ( saveStore.length > opts.storeMax ){
		  removeOldSaves( updateAutoSaveMenu );
		} else {
		  updateAutoSaveMenu();
		}
	};

	/**
	 * Update auto save dropdown menu
	 */
	let updateAutoSaveMenu = function(){
		localStorage.setItem( saveStoreKey, JSON.stringify( saveStore ) );
		let ulList = "";
		for ( let i = saveStore.length; i--; ){
			let newItemDate = DateTime.fromFormat( saveStore[ i ].replace( editorID + "_", "" ), "t" );
			let dateTitle = newItemDate.diffNow( "hours" ).hours < 1 ? newItemDate.toRelative() : newItemDate.toFormat( "MM/dd/yyyy h:mm a" );
			ulList += "<li><a href=\"javascript:void(0)\" data-id=\"" + saveStore[ i ] + "\">" + dateTitle +"</a></li>";
		}

		// No records
		if ( !saveStore.length ){
			ulList = "<li><a href=\"javascript:void(0)\">No Autosaves, type something :)</a></li>";
		}

		// Add records
		$( "#" + ddMenuID ).find( ".autoSaveMenu" ).html( ulList );
	};

	/**
	 * Start the autosave timer
	 * @param  {object} event The JS event object
	 */
	let startTimer = function( event ){
	  	if ( timer ){ clearTimeout( timer ); }
	  	timer = setTimeout( onTimer, opts.timeout, event );
	};

	/**
	 * Autosave Command
	 * @param  {object} event The JS event object
	 */
	let onTimer = function( event ){
		if ( savingActive ) {
		  startTimer( event );
		} else {
			savingActive = true;
			let autoSaveKey = editorID + "_" + Date.now();
			// Store it
			localStorage.setItem( autoSaveKey, compressToUTF16( getEditorContent() ) );
			// Add to items
			addToStore( autoSaveKey );
			// mark as done
			savingActive = false;
		}
	};

	/**
	 * Load content back into editor
	 * @param  {string} contentID The content ID to load from local storage
	 */
	let loadContent = function( contentID ){
		let content = localStorage.getItem( contentID );
		setEditorContent( "content", decompressFromUTF16( content ) );
		if ( timer ){ clearTimeout( timer ); }
	};

	// Register change event for auto saving
	getContentEditor().on( "change", startTimer );

	// Load Previous AutoLoad when selected from the Dropdown menu
	$( "#" + ddMenuID ).on( "click", "li > a", function( evt ){
		loadContent( $( evt.currentTarget ).data( "id" ) );
	} );

	// Update auto save menu
	updateAutoSaveMenu();
};
